#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import datetime
import logging

from .library import counters, notify_maintainer

####################################################################################################
def check_marks(ports, limits):
    """ Checks the existence of special marks variables (ie. BROKEN, etc.) in Makefiles """
    for name, port in ports.items():
        today = datetime.datetime.now(datetime.timezone.utc)

        if "BROKEN" in port:
            if port["Last modification"] < today - datetime.timedelta(days=limits["BROKEN since"]):
                logging.warning("BROKEN mark '%s' for port %s", port["BROKEN"], name)
                counters["Marked as BROKEN for too long"] += 1
                notify_maintainer(port["maintainer"], "Marked as BROKEN for too long", name)
            else:
                logging.info("BROKEN mark '%s' for port %s", port["BROKEN"], name)
                counters["Marked as BROKEN"] += 1
                notify_maintainer(port["maintainer"], "Marked as BROKEN", name)

        if "DEPRECATED" in port:
            if port["Last modification"] < today - datetime.timedelta(days=limits["DEPRECATED since"]):
                logging.warning("DEPRECATED mark '%s' for port %s", port["DEPRECATED"], name)
                counters["Marked as DEPRECATED for too long"] += 1
                notify_maintainer(port["maintainer"], "Marked as DEPRECATED for too long", name)
            else:
                logging.info("DEPRECATED mark '%s' for port %s", port["DEPRECATED"], name)
                counters["Marked as DEPRECATED"] += 1
                notify_maintainer(port["maintainer"], "Marked as DEPRECATED", name)

        if "FORBIDDEN" in port:
            if port["Last modification"] < today - datetime.timedelta(days=limits["FORBIDDEN since"]):
                logging.warning("FORBIDDEN mark '%s' for port %s", port["FORBIDDEN"], name)
                counters["Marked as FORBIDDEN for too long"] += 1
                notify_maintainer(port["maintainer"], "Marked as FORBIDDEN for too long", name)
            else:
                logging.info("FORBIDDEN mark '%s' for port %s", port["FORBIDDEN"], name)
                counters["Marked as FORBIDDEN"] += 1
                notify_maintainer(port["maintainer"], "Marked as FORBIDDEN", name)

        if "IGNORE" in port:
            logging.info("IGNORE mark '%s' for port %s", port["IGNORE"], name)
            counters["Marked as IGNORE"] += 1
            notify_maintainer(port["maintainer"], "Containing an IGNORE mark", name)

        if "RESTRICTED" in port:
            logging.info("RESTRICTED mark '%s' for port %s", port["RESTRICTED"], name)
            counters["Marked as RESTRICTED"] += 1
            notify_maintainer(port["maintainer"], "Marked as RESTRICTED", name)

        if "EXPIRATION_DATE" in port:
            logging.warning("EXPIRATION_DATE mark '%s' for port %s", port["EXPIRATION_DATE"], name)
            counters["Marked with EXPIRATION_DATE"] += 1
            notify_maintainer(port["maintainer"], "Marked with EXPIRATION_DATE", name)

    logging.info("Found %d ports marked as BROKEN", counters["Marked as BROKEN"])
    logging.info("Found %d ports marked as BROKEN for too long", counters["Marked as BROKEN for too long"])
    logging.info("Found %d ports marked as DEPRECATED", counters["Marked as DEPRECATED"])
    logging.info("Found %d ports marked as DEPRECATED for too long", counters["Marked as DEPRECATED for too long"])
    logging.info("Found %d ports marked as FORBIDDEN", counters["Marked as FORBIDDEN"])
    logging.info("Found %d ports marked as FORBIDDEN for too long", counters["Marked as FORBIDDEN for too long"])
    logging.info("Found %d ports marked as IGNORE", counters["Marked as IGNORE"])
    logging.info("Found %d ports marked as RESTRICTED", counters["Marked as RESTRICTED"])
    logging.info("Found %d ports marked with EXPIRATION_DATE", counters["Marked with EXPIRATION_DATE"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import datetime
import logging
import os
import re
import sys

import libpnu

from .library import counters, notify_maintainer

####################################################################################################
def load_freebsd_ports_dict(ports_dir):
    """ Returns a dictionary of FreeBSD ports """
    ports = {}

    # Are we running on FreeBSD?
    operating_system = sys.platform
    if not operating_system.startswith("freebsd"):
        raise SystemError

    # On which version?
    os_version = operating_system.replace("freebsd", "")

    # Is the ports index installed?
    ports_index = ports_dir + os.sep + "INDEX-" + os_version
    if not os.path.isfile(ports_index):
        raise FileNotFoundError

    # Loading the ports index:
    with open(ports_index, encoding='utf-8', errors='ignore') as file:
        lines = file.read().splitlines()

    for line in lines:
        # The file format is described at: https://wiki.freebsd.org/Ports/INDEX
        fields = line.split('|')
        if len(fields) != 13:
            logging.error("Ports index line '%s' has %d fields instead of the expected 13. Line ignored", line, len(fields))
        elif fields[0] in ports:
            logging.error("Ports index line '%s' refers to a duplicate distribution-name. Line ignored", line)
        else:
            ports[fields[0]] = \
                {
                    "port-path": fields[1],
                    "installation-prefix": fields[2],
                    "comment": fields[3],
                    "description-file": fields[4],
                    "maintainer": fields[5].lower(),
                    "categories": fields[6],
                    "extract-depends": fields[7],
                    "patch-depends": fields[8],
                    "www-site": fields[9],
                    "fetch-depends": fields[10],
                    "build-depends": fields[11],
                    "run-depends": fields[12],
                }

    counters["FreeBSD ports"] = len(ports)
    logging.info("Loaded %d ports from the FreeBSD Ports INDEX file", len(ports))
    return ports


####################################################################################################
def filter_ports(ports, selected_categories, selected_maintainers, selected_ports):
    """ Filters the list of ports to the specified categories AND maintainers"""
    if selected_categories or selected_maintainers or selected_ports:
        all_ports = " ".join(ports.keys())
        for port in all_ports.split():
            if selected_maintainers:
                if ports[port]["maintainer"] not in selected_maintainers:
                    del ports[port]
                    continue
            if selected_categories:
                match = False
                for category in ports[port]["categories"].split():
                    if category in selected_categories:
                        match = True
                        break
                if not match:
                    del ports[port]
                    continue
            if selected_ports:
                port_id = re.sub(r".*/", "", ports[port]["port-path"])
                if port_id not in selected_ports:
                    del ports[port]

    counters["Selected ports"] = len(ports)
    logging.info("Selected %d ports", len(ports))
    return ports


####################################################################################################
def update_with_makefiles(ports, ports_dir):
    """ Loads selected part of port's Makefiles for cross-checking things """
    for name, port in ports.items():
        # Use the PORTSDIR we have been told to, rather than the system's one
        port_path = port["port-path"].replace("/usr/ports", ports_dir)
        if not os.path.isdir(port_path):
            continue

        port_makefile = port_path + os.sep + 'Makefile'
        if not os.path.isfile(port_makefile):
            logging.error("Nonexistent Makefile for port %s", name)
            counters["Nonexistent Makefile"] += 1
            notify_maintainer(port["maintainer"], "Nonexistent Makefile", name)
        else:
            # Getting the port last modification datetime:
            ports[name]["Last modification"] = datetime.datetime.fromtimestamp(os.path.getmtime(port_makefile)).replace(tzinfo=datetime.timezone.utc)

            lines = libpnu.load_strings_from_file(port_makefile)

            for line in lines:
                group = re.match(r"^([A-Z_]+)=[ 	]*(.*)", line)
                if group is not None: # Makefile variable
                    ports[name][group[1]] = group[2]

    logging.info("Found %d ports with nonexistent Makefile", counters["Nonexistent Makefile"])
    return ports



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from .library import counters, notify_maintainer

####################################################################################################
def check_installation_prefix(ports):
    """ Checks the installation-prefix field usualness """
    for name, port in ports.items():
        if port["installation-prefix"] == "/usr/local":
            pass
        elif port["installation-prefix"] == "/compat/linux" and name.startswith("linux"):
            pass
        elif port["installation-prefix"] == "/usr/local/FreeBSD_ARM64" and "-aarch64-" in name:
            pass
        elif port["installation-prefix"].startswith("/usr/local/android") and "droid" in name:
            pass
        elif port["installation-prefix"] == "/var/qmail" and "qmail" in name or name.startswith("queue-fix"):
            pass
        elif port["installation-prefix"] == "/usr" and name.startswith("global-tz-") or name.startswith("zoneinfo-"):
            pass
        else:
            logging.warning("Unusual installation-prefix '%s' for port %s", port["installation-prefix"], name)
            counters["Unusual installation-prefix"] += 1
            notify_maintainer(port["maintainer"], "Unusual installation-prefix", name)

    logging.info("Found %d ports with unusual installation-prefix", counters["Unusual installation-prefix"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import textwrap

from .library import notifications

CSV_FILE_SEPARATOR = ';' # don't use ',' which can appear in port's names

####################################################################################################
def show_notifications():
    """ Pretty prints notifications """
    sorted_notifications = dict(sorted(notifications.items()))
    print("\nIssues per maintainer:")
    for maintainer, details in sorted_notifications.items():
        print(f"  {maintainer}:")
        for issue, ports in details.items():
            print(f"    {issue}:")
            all_ports = " ".join(ports)
            for line in textwrap.wrap(all_ports, width=74, break_on_hyphens=False):
                print(f"      {line}")
        print()


####################################################################################################
def output_notifications(filename):
    """ Output notifications in a CSV file """
    sorted_notifications = dict(sorted(notifications.items()))
    try:
        file = open(filename, "w",  encoding='utf-8')
    except Exception:
        logging.error("Unable to save per-maintainer output to file '%s'", filename)
    else:
        with file:
            print("MAINTAINER" + CSV_FILE_SEPARATOR + "ISSUE" + CSV_FILE_SEPARATOR + "PORT", file=file)
            for maintainer, details in sorted_notifications.items():
                for issue, ports in details.items():
                    for port in ports:
                        print(maintainer + CSV_FILE_SEPARATOR + issue + CSV_FILE_SEPARATOR + port, file=file)



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
Contributor: Emanuel Haupt
"""

import getopt
import logging
import os
import sys

import libpnu

from .load_data import load_freebsd_ports_dict, filter_ports, update_with_makefiles
from .load_config import generate_config, load_config
from .check_port_path import check_port_path
from .check_installation_prefix import check_installation_prefix
from .check_comment import check_comment
from .check_description_file import check_description_file
from .check_plist import check_plist
from .check_maintainer import check_maintainer
from .check_categories import check_categories
from .check_www_site import check_www_site
from .check_marks import check_marks
from .check_unchanging_ports import check_unchanging_ports
from .check_vulnerabilities import check_vulnerabilities
from .show_categories import show_categories
from .show_maintainers import show_maintainers
from .show_notifications import show_notifications, output_notifications
from .show_summary import show_summary


# Version string used by the what(1) and ident(1) commands:
ID = "@(#) $Id: portstreelint - FreeBSD ports tree lint v1.3.0 (April 1st, 2024) by Hubert Tournier $"

# Default parameters. Can be overcome by command line options:
parameters = {
    "Load config": True,
    "INI filename": "",
    "Ports dir": "/usr/ports",
    "CSV filename": "",
    "Show categories": False,
    "Show maintainers": False,

    "Checks": {
        "port-path": True,
		"installation-prefix": True,
		"comment": True,
		"description-file": True,
		"plist": True,
		"maintainer": True,
		"categories": True,
		"www-site": True,
        "Hostnames": False,
        "URL": False,
		"Marks": True,
		"Unchanging ports": True,
		"Vulnerabilities": True,
    },
    "Limits": {
        "PLIST abuse": 7, # entries
        "BROKEN since": 6 * 30, # days
        "FORBIDDEN since": 3 * 30, # days
        "DEPRECATED since": 6 * 30, # days
        "Unchanged since": 3 * 365, # days
    },
    "Selections": {
        "Categories": [],
        "Maintainers": [],
        "Ports": [],
    },
    "Exclusions": {
        "Vulnerabilities": [],
    },
}


####################################################################################################
def _display_help():
    """ Display usage and help """
    #pylint: disable=C0301
    print("usage: portstreelint [--nocfg|-n] [--gencfg|-g FILE]", file=sys.stderr)
    print("        [--tree|-t DIR] [--show-cat|-C] [--show-mnt|-M]", file=sys.stderr)
    print("        [--cat|-c LIST] [--mnt|-m LIST] [--port|-p LIST] [--plist NUM]", file=sys.stderr)
    print("        [--broken NUM] [--deprecated NUM] [--forbidden NUM] [--unchanged NUM]", file=sys.stderr)
    print("        [--check-host|-h] [--check-url|-u] [--output|-o FILE]", file=sys.stderr)
    print("        [--debug] [--help|-?] [--info] [--version] [--]", file=sys.stderr)
    print("  ------------------  -------------------------------------------------------", file=sys.stderr)
    print("  --nocfg|-n          Don't use the configuration file", file=sys.stderr)
    print("  --gencfg|-g FILE    Generate a default configuration file in FILE", file=sys.stderr)
    print("  --show-cat|-C       Show categories with ports count", file=sys.stderr)
    print("  --show-mnt|-M       Show maintainers with ports count", file=sys.stderr)
    print("  --cat|-c LIST       Select only the comma-separated categories in LIST", file=sys.stderr)
    print("  --mnt|-m LIST       Select only the comma-separated maintainers in LIST", file=sys.stderr)
    print("  --port|-p LIST      Select only the comma-separated ports in LIST", file=sys.stderr)
    print("  --tree|-t DIR       Set ports directory (default=/usr/ports)", file=sys.stderr)
    print("  --plist NUM         Set PLIST_FILES abuse to NUM files", file=sys.stderr)
    print("  --broken NUM        Set BROKEN since to NUM days", file=sys.stderr)
    print("  --deprecated NUM    Set DEPRECATED since to NUM days", file=sys.stderr)
    print("  --forbidden NUM     Set FORBIDDEN since to NUM days", file=sys.stderr)
    print("  --unchanged NUM     Set Unchanged since to NUM days", file=sys.stderr)
    print("  --check-host|-h     Enable checking hostname resolution (long!)", file=sys.stderr)
    print("  --check-url|-u      Enable checking URL (very long!)", file=sys.stderr)
    print("  --output|-o FILE    Enable per-maintainer CSV output to FILE", file=sys.stderr)
    print("  --debug             Enable logging at debug level", file=sys.stderr)
    print("  --info              Enable logging at info level", file=sys.stderr)
    print("  --version           Print version and exit", file=sys.stderr)
    print("  --help|-?           Print usage and this help message and exit", file=sys.stderr)
    print("  --                  Options processing terminator", file=sys.stderr)
    print(file=sys.stderr)
    #pylint: enable=C0301


####################################################################################################
def _process_environment_variables():
    """ Process environment variables """
    #pylint: disable=C0103, W0602
    global parameters
    #pylint: enable=C0103, W0602

    if "PORTSTREELINT_DEBUG" in os.environ:
        logging.disable(logging.NOTSET)

    if "PORTSDIR" in os.environ:
        if os.environ["PORTSDIR"].endswith(os.sep):
            parameters["Ports dir"] = os.environ["PORTSDIR"][:-1]
        else:
            parameters["Ports dir"] = os.environ["PORTSDIR"]

    logging.debug("_process_environment_variables(): parameters:")
    logging.debug(parameters)


####################################################################################################
def _process_command_line():
    """ Process command line options """
    #pylint: disable=C0103, W0602
    global parameters
    #pylint: enable=C0103, W0602

    # option letters followed by : expect an argument
    # same for option strings followed by =
    character_options = "CMc:g:hm:no:p:t:u?"
    string_options = [
        "broken=",
        "cat=",
        "check-host",
        "check-url",
        "debug",
        "deprecated=",
        "forbidden=",
        "gencfg=",
        "help",
        "info",
        "mnt=",
        "nocfg",
        "output=",
        "port=",
        "plist=",
        "show-cat",
        "show-mnt",
        "tree=",
        "unchanged=",
        "version",
    ]

    try:
        options, remaining_arguments = getopt.getopt(
            sys.argv[1:], character_options, string_options
        )
    except getopt.GetoptError as error:
        logging.critical("Syntax error: %s", error)
        _display_help()
        sys.exit(1)

    for option, argument in options:
        if option == "--debug":
            logging.disable(logging.NOTSET)

        elif option in ("--help", "-?"):
            _display_help()
            sys.exit(0)

        elif option == "--info":
            logging.disable(logging.DEBUG)

        elif option == "--version":
            print(ID.replace("@(" + "#)" + " $" + "Id" + ": ", "").replace(" $", ""))
            sys.exit(0)

        elif option == "--broken":
            try:
                parameters["Checks"]["BROKEN since"] = int(argument)
            except ValueError:
                logging.critical("Syntax error: expecting a number of days after the broken option")
                sys.exit(1)
            if parameters["Checks"]["BROKEN since"] < 30:
                logging.critical("The number of days after the broken option must be >= 30")
                sys.exit(1)

        elif option in ("--cat", "-c"):
            parameters["Selections"]["Categories"] = argument.lower().split(",")

        elif option in ("--check-host", "-h"):
            parameters["Checks"]["Hostnames"] = True

        elif option in ("--check-url", "-u"):
            parameters["Checks"]["Hostnames"] = True
            parameters["Checks"]["URL"] = True

        elif option == "--deprecated":
            try:
                parameters["Checks"]["DEPRECATED since"] = int(argument)
            except ValueError:
                logging.critical("Syntax error: expecting a number of days after the deprecated option")
                sys.exit(1)
            if parameters["Checks"]["DEPRECATED since"] < 30:
                logging.critical("The number of days after the deprecated option must be >= 30")
                sys.exit(1)

        elif option == "--forbidden":
            try:
                parameters["Checks"]["FORBIDDEN since"] = int(argument)
            except ValueError:
                logging.critical("Syntax error: expecting a number of days after the forbidden option")
                sys.exit(1)
            if parameters["Checks"]["FORBIDDEN since"] < 30:
                logging.critical("The number of days after the forbidden option must be >= 30")
                sys.exit(1)

        elif option in ("--gencfg", "-g"):
            if not os.path.exists(argument):
                parameters["INI filename"] = argument
            else:
                logging.critical("--gencfg|-g argument cannot be an existing filesystem item")
                sys.exit(1)

        elif option in ("--mnt", "-m"):
            maintainers = argument.lower().split(",")
            parameters["Selections"]["Maintainers"] = [m if '@' in m else f"{m}@freebsd.org" for m in maintainers]

        elif option in ("--nocfg", "-n"):
            parameters["Load config"] = False

        elif option in ("--output", "-o"):
            parameters["CSV filename"] = argument

        elif option == "--plist":
            try:
                parameters["Checks"]["PLIST abuse"] = int(argument)
            except ValueError:
                logging.critical("Syntax error: expecting a number of files after the plist option")
                sys.exit(1)
            if parameters["Checks"]["PLIST abuse"] < 3:
                logging.critical("The number of files after the plist option must be >= 3")
                sys.exit(1)

        elif option in ("--port", "-p"):
            parameters["Selections"]["Ports"] = argument.split(",")

        elif option in ("--show-cat", "-C"):
            parameters["Show categories"] = True
            parameters["Show maintainers"] = False

        elif option in ("--show-mnt", "-M"):
            parameters["Show maintainers"] = True
            parameters["Show categories"] = False

        elif option in ("--tree", "-t"):
            if os.path.isdir(argument):
                if argument.endswith(os.sep):
                    parameters["Ports dir"] = argument[:-1]
                else:
                    parameters["Ports dir"] = argument
            else:
                logging.critical("--tree|-t argument must be a directory name")
                sys.exit(1)

        elif option == "--unchanged":
            try:
                parameters["Checks"]["Unchanged since"] = int(argument)
            except ValueError:
                logging.critical("Syntax error: expecting a number of days after the unchanged option")
                sys.exit(1)
            if parameters["Checks"]["Unchanged since"] < 30:
                logging.critical("The number of days after the unchanged option must be >= 30")
                sys.exit(1)

    logging.debug("_process_command_line(): parameters:")
    logging.debug(parameters)
    logging.debug("_process_command_line(): remaining_arguments:")
    logging.debug(remaining_arguments)

    return remaining_arguments


####################################################################################################
def main():
    """ The program's main entry point """
    #pylint: disable=C0103, W0602
    global parameters
    #pylint: enable=C0103, W0602

    program_name = os.path.basename(sys.argv[0])

    libpnu.initialize_debugging(program_name)
    libpnu.handle_interrupt_signals(libpnu.interrupt_handler_function)

    if parameters["Load config"]:
        parameters = load_config(parameters)
    _process_environment_variables()
    _ = _process_command_line()

    # Load the FreeBSD ports INDEX file
    # and verify structural integrity (ie: 13 pipe-separated fields)
    # and unicity of distribution-names
    try:
        ports = load_freebsd_ports_dict(parameters["Ports dir"])
    except SystemError:
        logging.critical("This program will only run on a FreeBSD operating system")
        sys.exit(1)
    except FileNotFoundError:
        logging.critical("The ports tree is missing. Please install and update it as root ('portsnap fetch extract')")
        sys.exit(1)

    if parameters["INI filename"]:
        # Only generate a configuration file
        generate_config(parameters)
    elif parameters["Show categories"]:
        # Only show port categories with count of associated ports
        show_categories(ports)
    elif parameters["Show maintainers"]:
        # Only show port maintainers with count of associated ports
        show_maintainers(ports)
    else:
        ports = filter_ports(
            ports,
            parameters["Selections"]["Categories"],
            parameters["Selections"]["Maintainers"],
            parameters["Selections"]["Ports"]
        )

        # Check the existence of port Makefile and load its variables
        ports = update_with_makefiles(ports, parameters["Ports dir"])

        # Check the existence of port-path
        if parameters["Checks"]["port-path"]:
            check_port_path(ports, parameters["Ports dir"])

        # Check unusual installation-prefix
        if parameters["Checks"]["installation-prefix"]:
            check_installation_prefix(ports)

        # Cross-check comment identity between Index and Makefile
        if parameters["Checks"]["comment"]:
            check_comment(ports)

        # Check the existence of description-file
        if parameters["Checks"]["description-file"]:
            check_description_file(ports, parameters["Ports dir"])

        # Check the package list
        if parameters["Checks"]["plist"]:
            check_plist(ports, parameters["Limits"]["PLIST abuse"], parameters["Ports dir"])

        # Cross-check maintainer identity between Index and Makefile
        if parameters["Checks"]["maintainer"]:
            check_maintainer(ports)

        # Cross-check categories identity between Index and Makefile
        # and belonging to the official list of categories
        if parameters["Checks"]["categories"]:
            check_categories(ports)

        # Check that www-site is not empty, that the hostnames exist,
        # that the URL is accessible, and identity between Index and Makefile
        if parameters["Checks"]["www-site"]:
            check_www_site(ports, parameters["Checks"]["Hostnames"], parameters["Checks"]["URL"])

        # Check the existence of marks variables (ie. BROKEN, etc.) in Makefiles
        if parameters["Checks"]["Marks"]:
            check_marks(ports, parameters["Limits"])

        # Check static ports
        if parameters["Checks"]["Unchanging ports"]:
            check_unchanging_ports(ports, parameters["Limits"]["Unchanged since"])

        # Check vulnerabilities
        if parameters["Checks"]["Vulnerabilities"]:
            check_vulnerabilities(ports, parameters["Exclusions"]["Vulnerabilities"])

        # Print results per maintainer
        show_notifications()

        # Output per maintainer results in a CSV file
        if parameters["CSV filename"]:
            output_notifications(parameters["CSV filename"])

        # Print summary of findings
        show_summary(parameters["Limits"])

    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import datetime
import logging

from .library import counters, notify_maintainer

####################################################################################################
def check_unchanging_ports(ports, unchanged_days):
    """ Checks if the port has been unmodified for too long """
    for name, port in ports.items():
        today = datetime.datetime.now(datetime.timezone.utc)
        if "Last modification" in port:
            if port["Last modification"] < today - datetime.timedelta(days=unchanged_days):
                logging.info("No modification since more than %d days for port %s", unchanged_days, name)
                counters["Unchanged for a long time"] += 1
                notify_maintainer(port["maintainer"], "Unchanged for a long time", name)

    logging.info("Found %d ports unchanged for a long time", counters["Unchanged for a long time"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import configparser
import logging
import os

import libpnu


####################################################################################################
def generate_config(parameters):
    """ Generates a configuration file with the default + environment + command line parameters """

    content = f"""# PortsTreeLint configuration file

# Check the following URL for the file format:
# https://docs.python.org/3/library/configparser.html#supported-ini-file-structure

[params]
ports_dir = {parameters['Ports dir']}
csv_filename =
debug_level = warning
#debug_level = info
#debug_level = debug

[checks]
port_path = {parameters['Checks']['port-path']}
installation_prefix = {parameters['Checks']['installation-prefix']}
comment = {parameters['Checks']['comment']}
description_file = {parameters['Checks']['description-file']}
plist = {parameters['Checks']['plist']}
maintainer = {parameters['Checks']['maintainer']}
categories = {parameters['Checks']['categories']}
www_site = {parameters['Checks']['www-site']}
hostnames = {parameters['Checks']['Hostnames']}
url = {parameters['Checks']['URL']}
marks = {parameters['Checks']['Marks']}
unchanging_ports = {parameters['Checks']['Unchanging ports']}
vulnerabilities = {parameters['Checks']['Vulnerabilities']}

[limits]
# number of entries:
plist_abuse = {parameters['Limits']['PLIST abuse']}
# number of days:
broken_since = {parameters['Limits']['BROKEN since']}
forbidden_since = {parameters['Limits']['FORBIDDEN since']}
deprecated_since = {parameters['Limits']['DEPRECATED since']}
unchanged_since = {parameters['Limits']['Unchanged since']}

[selections]
# (multilines) lists of space separated values:
categories =
maintainers =
ports =

[exclusions]
# (multilines) lists of space separated Vulnerabilities IDs:
vulnerabilities =
"""

    with open(parameters["INI filename"], "w", encoding="utf-8") as file:
        file.write(content)


####################################################################################################
def _string2bool(string):
    """ Converts a string to a boolean without ValueError exception """
    if string.lower() in ("true", "yes", "on", "1"):
        return True

    # .getboolean() would generate an exception if string wasn't in "false", "no", "off", 0"...
    return False


####################################################################################################
def _string2int(string, default):
    """ Converts a string to an int or use a default value without ValueError exception """
    try:
        value = int(string)
    except ValueError:
        value = default

    return value


####################################################################################################
def load_config(parameters):
    """ Loads the user configuration file """
    home = libpnu.get_home_directory()
    if os.name == "nt":
        config_file = home + os.sep + "ptlint.ini"
    else:
        config_file = home + os.sep + ".ptlint"

    if os.path.isfile(config_file):
        config = configparser.ConfigParser()
        config.read(config_file)
        if "params" in config:
            if "ports_dir" in config["params"]:
                parameters["Ports dir"] = config["params"]["ports_dir"]
            if "csv_filename" in config["params"]:
                parameters["CSV filename"] = config["params"]["csv_filename"]
            if "debug_level" in config["params"]:
                if config["params"]["debug_level"] == "info":
                    logging.disable(logging.DEBUG)
                elif config["params"]["debug_level"] == "debug":
                    logging.disable(logging.NOTSET)
        if "checks" in config:
            if "port_path" in config["checks"]:
                parameters["Checks"]["port-path"] = _string2bool(config["checks"]["port_path"])
            if "installation_prefix" in config["checks"]:
                parameters["Checks"]["installation-prefix"] = _string2bool(config["checks"]["installation_prefix"])
            if "comment" in config["checks"]:
                parameters["Checks"]["comment"] = _string2bool(config["checks"]["comment"])
            if "description_file" in config["checks"]:
                parameters["Checks"]["description-file"] = _string2bool(config["checks"]["description_file"])
            if "plist" in config["checks"]:
                parameters["Checks"]["plist"] = _string2bool(config["checks"]["plist"])
            if "maintainer" in config["checks"]:
                parameters["Checks"]["maintainer"] = _string2bool(config["checks"]["maintainer"])
            if "categories" in config["checks"]:
                parameters["Checks"]["categories"] = _string2bool(config["checks"]["categories"])
            if "www_site" in config["checks"]:
                parameters["Checks"]["www-site"] = _string2bool(config["checks"]["www_site"])
            if "hostnames" in config["checks"]:
                parameters["Checks"]["Hostnames"] = _string2bool(config["checks"]["hostnames"])
            if "url" in config["checks"]:
                parameters["Checks"]["URL"] = _string2bool(config["checks"]["url"])
            if "marks" in config["checks"]:
                parameters["Checks"]["Marks"] = _string2bool(config["checks"]["marks"])
            if "unchanging_ports" in config["checks"]:
                parameters["Checks"]["Unchanging ports"] = _string2bool(config["checks"]["unchanging_ports"])
            if "vulnerabilities" in config["checks"]:
                parameters["Checks"]["Vulnerabilities"] = _string2bool(config["checks"]["vulnerabilities"])
        if "limits" in config:
            if "plist_abuse" in config["limits"]:
                parameters["Limits"]["PLIST abuse"] = _string2int(config["limits"]["plist_abuse"], parameters["Limits"]["PLIST abuse"])
            if "broken_since" in config["limits"]:
                parameters["Limits"]["BROKEN since"] = _string2int(config["limits"]["broken_since"], parameters["Limits"]["BROKEN since"])
            if "forbidden_since" in config["limits"]:
                parameters["Limits"]["FORBIDDEN since"] = _string2int(config["limits"]["forbidden_since"], parameters["Limits"]["FORBIDDEN since"])
            if "deprecated_since" in config["limits"]:
                parameters["Limits"]["DEPRECATED since"] = _string2int(config["limits"]["deprecated_since"], parameters["Limits"]["DEPRECATED since"])
            if "unchanged_since" in config["limits"]:
                parameters["Limits"]["Unchanged since"] = _string2int(config["limits"]["unchanged_since"], parameters["Limits"]["Unchanged since"])
        if "selections" in config:
            if "categories" in config["selections"]:
                parameters["Selections"]["Categories"] = config["selections"]["categories"].split()
            if "maintainers" in config["selections"]:
                parameters["Selections"]["Maintainers"] = [m if '@' in m else f"{m}@freebsd.org" for m in config["selections"]["maintainers"].split()]
            if "ports" in config["selections"]:
                parameters["Selections"]["Ports"] = config["selections"]["ports"].split()
        if "exclusions" in config:
            if "vulnerabilities" in config["exclusions"]:
                parameters["Exclusions"]["Vulnerabilities"] = config["exclusions"]["vulnerabilities"].split()

    return parameters



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from .library import counters, notify_maintainer

####################################################################################################
def check_maintainer(ports):
    """ Cross-checks the maintainer field with the Makefile and compliance with rules
    Rules at https://docs.freebsd.org/en/books/porters-handbook/makefiles/#makefile-maintainer
    """
    for name, port in ports.items():
        if "MAINTAINER" in port:
            if '$' in port["MAINTAINER"]:
                continue # don't try to resolve embedded variables. Ignore check

            if port["maintainer"] != port["MAINTAINER"].lower():
                logging.error("Diverging maintainers between Index and Makefile for port %s", name)
                logging.error("... Index:maintainer    '%s'", port["maintainer"])
                logging.error("... Makefile:MAINTAINER '%s'", port["MAINTAINER"])
                counters["Diverging maintainers"] += 1
                notify_maintainer(port["maintainer"], "Diverging maintainers", name)
                notify_maintainer(port["MAINTAINER"], "Diverging maintainers", name)

    logging.info("Found %d ports with diverging maintainers", counters["Diverging maintainers"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import re

import vuxml

from .library import counters, notify_maintainer

####################################################################################################
def _debug_versions(port):
    """ Prints debugging info about a port version """
    pkgnameprefix = ""
    if "PKGNAMEPREFIX" in port:
        pkgnameprefix = port["PKGNAMEPREFIX"]
    portname = ""
    if "PORTNAME" in port:
        portname = port["PORTNAME"]
    pkgnamesuffix = ""
    if "PKGNAMESUFFIX" in port:
        pkgnamesuffix = port["PKGNAMESUFFIX"]
    portversion = ""
    if "PORTVERSION" in port:
        portversion = port["PORTVERSION"]
    portrevision = ""
    if "PORTREVISION" in port:
        portrevision = port["PORTREVISION"]
    portepoch = ""
    if "PORTEPOCH" in port:
        portepoch = port["PORTEPOCH"]

    distversionprefix = ""
    if "DISTVERSIONPREFIX" in port:
        distversionprefix = port["DISTVERSIONPREFIX"]
    distversion = ""
    if "DISTVERSION" in port:
        distversion = port["DISTVERSION"]
    distversionsuffix = ""
    if "DISTVERSIONSUFFIX" in port:
        distversionsuffix = port["DISTVERSIONSUFFIX"]

    logging.debug("  PKGNAMEPREFIX=%s", pkgnameprefix)
    logging.debug("  PORTNAME=%s", portname)
    logging.debug("  PKGNAMESUFFIX=%s", pkgnamesuffix)
    logging.debug("  PORTVERSION=%s", portversion)
    logging.debug("  PORTREVISION=%s", portrevision)
    logging.debug("  PORTEPOCH=%s", portepoch)
    logging.debug("  DISTVERSIONPREFIX=%s", distversionprefix)
    logging.debug("  DISTVERSION=%s", distversion)
    logging.debug("  DISTVERSIONSUFFIX=%s", distversionsuffix)


####################################################################################################
def check_vulnerabilities(ports, excluded_vulnerabilities):
    """ Checks if the port has vulnerabilities reported in VuXML """
    vulns = vuxml.load_vuxml()
    logging.info("Loaded %d vulnerabilities from the FreeBSD VuXML files", len(vulns))

    for name, port in ports.items():
        portname = ""
        if "PORTNAME" in port:
            portname = port["PORTNAME"]
        else:
            logging.debug("No PORTNAME for port %s", name)

        version = ""

        portversion = ""
        if "PORTVERSION" in port:
            portversion = port["PORTVERSION"]
        distversion = ""
        if "DISTVERSION" in port:
            distversion = port["DISTVERSION"]
        portrevision = ""
        if "PORTREVISION" in port:
            portrevision = port["PORTREVISION"]
        portepoch = ""
        if "PORTEPOCH" in port:
            portepoch = port["PORTEPOCH"]

        if portversion and distversion:
            logging.error("Both PORTVERSION and DISTVERSION for port %s", name)
            counters["Both PORTVERSION and DISTVERSION"] += 1
            notify_maintainer(port["maintainer"], "Both PORTVERSION and DISTVERSION", name)

        if not portversion and not distversion:
            logging.debug("No PORTVERSION and DISTVERSION for port %s", name)

        if portversion and '$' not in portversion:
            version = portversion
            if portrevision:
                version += '_' + portrevision

        # Try to figure out ourselves from the port name:
        if not portname or not version:
            version = name

            if portepoch:
                version = re.sub(r"," + portepoch + "$", "", version)
            elif ',' in version:
                version = re.sub(r",[0-9]+$", "", version)
                logging.debug("Port epoch without PORTEPOCH for port %s", name)

            if not portrevision and '_' in version:
                logging.debug("Port revision without PORTREVISION for port %s", name)

            group = re.match(r"^(.*)-([^-]+)$", version)
            if group is not None:
                if not portname:
                    portname = group[1]
                version = group[2]
                logging.debug("portname='%s' version='%s' assumed for port %s", portname, version, name)
            else:
                logging.warning("Unable to get version for port %s. Skipping vulnerability check!", name)
                _debug_versions(port)
                counters["Skipped vulnerability checks"] += 1
                continue

        try:
            vids = vuxml.search_vulns_by_package(vulns, portname, version)
        except Exception as error:
            logging.warning('Encountered "%s" while searching vulnerabilities for port %s. Skipping vulnerability check', error, name)
            counters["Skipped vulnerability checks"] += 1
            continue

        for vid in excluded_vulnerabilities:
            if vid in vids:
                vids.remove(vid)
                logging.debug("Discarded false-positive VuXML vulnerability '%s' for port %s", vid, name)

        for vid in vids:
            logging.warning("Found VuXML vulnerability '%s' for port %s", vid, name)
            if not "FORBIDDEN" in port:
                logging.debug("Vulnerable port not marked as FORBIDDEN for port %s", name)

        if vids:
            counters["Vulnerable port version"] += 1
            notify_maintainer(port["maintainer"], "Vulnerable port", name)

    logging.info("Found %d ports with a vulnerable version", counters["Vulnerable port version"])
    logging.info("Skipped vulnerability check for %d ports", counters["Skipped vulnerability checks"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import textwrap

####################################################################################################
def show_maintainers(ports):
    """ Pretty prints port maintainers """
    maintainers = {}
    for _, port in ports.items():
        if port["maintainer"] in maintainers:
            maintainers[port["maintainer"]] += 1
        else:
            maintainers[port["maintainer"]] = 1

    sorted_maintainers = dict(sorted(maintainers.items()))
    all_maintainers = ""
    for maintainer, count in sorted_maintainers.items():
        all_maintainers += f"{maintainer}({count}), "
    all_maintainers = all_maintainers[:-2]

    print(f"Showing {len(maintainers)} maintainers with ports counts:\n")
    for line in textwrap.wrap(all_maintainers, width=80, break_on_hyphens=False):
        print(line)



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import os

from .library import counters, notify_maintainer

####################################################################################################
def check_port_path(ports, ports_dir):
    """ Checks the port-path field existence """
    for name, port in ports.items():
        # Use the PORTSDIR we have been told to, rather than the system's one
        port_path = port["port-path"].replace("/usr/ports", ports_dir)
        if not os.path.isdir(port_path):
            logging.error("Nonexistent port-path '%s' for port %s", port_path, name)
            counters["Nonexistent port-path"] += 1
            notify_maintainer(port["maintainer"], "Nonexistent port-path", name)

    logging.info("Found %d ports with nonexistent port-path", counters["Nonexistent port-path"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from .library import counters, notify_maintainer

OFFICIAL_CATEGORIES = [
    "accessibility", "afterstep", "arabic", "archivers", "astro", "audio", "benchmarks", "biology",
    "cad", "chinese", "comms", "converters", "databases", "deskutils", "devel", "dns", "docs",
    "editors", "education", "elisp", "emulators", "enlightenment", "finance", "french", "ftp",
    "games", "geography", "german", "gnome", "gnustep", "graphics", "hamradio", "haskell", "hebrew",
    "hungarian", "irc", "japanese", "java", "kde", "kde-applications", "kde-frameworks",
    "kde-plasma", "kld", "korean", "lang", "linux", "lisp", "mail", "mate", "math", "mbone", "misc",
    "multimedia", "net", "net-im", "net-mgmt", "net-p2p", "net-vpn", "news", "parallel", "pear",
    "perl5", "plan9", "polish", "ports-mgmt", "portuguese", "print", "python", "ruby", "rubygems",
    "russian", "scheme", "science", "security", "shells", "spanish", "sysutils", "tcl", "textproc",
    "tk", "ukrainian", "vietnamese", "wayland", "windowmaker", "www", "x11", "x11-clocks",
    "x11-drivers", "x11-fm", "x11-fonts", "x11-servers", "x11-themes", "x11-toolkits", "x11-wm",
    "xfce", "zope",
]

####################################################################################################
def check_categories(ports):
    """ Cross-checks the categories field with the Makefile and compliance with rules
    Rules at https://docs.freebsd.org/en/books/porters-handbook/makefiles/#makefile-categories-definition
    """
    for name, port in ports.items():
        for category in port["categories"].split():
            if category not in OFFICIAL_CATEGORIES:
                logging.warning("Unofficial category '%s' in Index for port %s", category, name)
                counters["Unofficial categories"] += 1
                notify_maintainer(port["maintainer"], "Unofficial category", name)

        if "CATEGORIES" in port:
            if '$' in port["CATEGORIES"]:
                continue # don't try to resolve embedded variables. Ignore check

            if port["categories"] != port["CATEGORIES"]:
                logging.error("Diverging categories between Index and Makefile for port %s", name)
                logging.error("... Index:categories    '%s'", port["categories"])
                logging.error("... Makefile:CATEGORIES '%s'", port["CATEGORIES"])
                counters["Diverging categories"] += 1
                notify_maintainer(port["maintainer"], "Diverging categories", name)

            for category in port["CATEGORIES"].split():
                if category not in OFFICIAL_CATEGORIES:
                    logging.warning("Unofficial category '%s' in Makefile for port %s", category, name)
                    counters["Unofficial categories"] += 1
                    notify_maintainer(port["maintainer"], "Unofficial category", name)

    logging.info("Found %d ports with unofficial categories", counters["Unofficial categories"])
    logging.info("Found %d ports with diverging categories", counters["Diverging categories"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from .library import counters, notify_maintainer

####################################################################################################
def check_comment(ports):
    """ Cross-checks the comment field with the Makefile and compliance with rules
    Rules at https://docs.freebsd.org/en/books/porters-handbook/makefiles/#makefile-comment
    """
    for name, port in ports.items():
        if len(port["comment"]) > 70:
            logging.warning("Over 70 characters comment for port %s", name)
            counters["Too long comments"] += 1
            notify_maintainer(port["maintainer"], "Too long comments", name)

        if 'a' <= port["comment"][0] <= 'z':
            logging.error("Uncapitalized comment for port %s", name)
            counters["Uncapitalized comments"] += 1
            notify_maintainer(port["maintainer"], "Uncapitalized comments", name)

        if port["comment"].endswith('.'):
            logging.error("Dot-ended comment for port %s", name)
            counters["Dot-ended comments"] += 1
            notify_maintainer(port["maintainer"], "Dot-ended comments", name)

        if "COMMENT" in port:
            if '$' in port["COMMENT"]:
                continue # don't try to resolve embedded variables. Ignore check

            # Do not take into escaping backslashes which are used inconsistently
            # in both fields
            if port["comment"].replace("\\", "") != port["COMMENT"].replace("\\", ""):
                logging.error("Diverging comments between Index and Makefile for port %s", name)
                logging.error("... Index:comment    '%s'", port["comment"])
                logging.error("... Makefile:COMMENT '%s'", port["COMMENT"])
                counters["Diverging comments"] += 1
                notify_maintainer(port["maintainer"], "Diverging comments", name)

    logging.info("Found %d ports with too long comments", counters["Too long comments"])
    logging.info("Found %d ports with uncapitalized comments", counters["Uncapitalized comments"])
    logging.info("Found %d ports with dot-ended comments", counters["Dot-ended comments"])
    logging.info("Found %d ports with diverging comments", counters["Diverging comments"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

from .library import counters

####################################################################################################
def _conditional_print(counter, message):
    """ Print a message if the counter is non zero """
    if counters[counter]:
        value = counters[counter]
        print(f"  {value} port{'' if value == 1 else 's'} {message}")


####################################################################################################
def show_summary(limits):
    """ Pretty prints a summary of findings """
    print(f'Selected {counters["Selected ports"]} ports out of {counters["FreeBSD ports"]} in the FreeBSD port tree, and found:')
    _conditional_print("Nonexistent port-path", "with non existent port-path")
    _conditional_print("Nonexistent Makefile", "without Makefile")
    _conditional_print("Unusual installation-prefix", "with unusual installation-prefix (warning)")
    _conditional_print("Too long comments", "with a comment string exceeding 70 characters (warning)")
    _conditional_print("Uncapitalized comments", "with an uncapitalized comment")
    _conditional_print("Dot-ended comments", "comment ending with a dot")
    _conditional_print("Diverging comments", "with a comment different between the Index and Makefile")
    _conditional_print("Nonexistent description-file", "with non existent description-file")
    _conditional_print("URL ending description-file", "with URL ending description-file")
    _conditional_print("description-file same as comment", "with description-file identical to comment")
    _conditional_print("Too short description-file", "with description-file no longer than comment")
    _conditional_print("Nonexistent pkg-plist", "without pkg-plist/PLIST_FILES/PLIST/PLIST_SUB (info)")
    _conditional_print("PLIST_FILES abuse", f"abusing PLIST_FILES with more than {limits['PLIST abuse'] - 1} entries (warning)")
    _conditional_print("Diverging maintainers", "with a maintainer different between the Index and Makefile")
    _conditional_print("Unofficial categories", "referring to unofficial categories (warning)")
    _conditional_print("Diverging categories", "with categories different between the Index and Makefile")
    _conditional_print("Empty www-site", "with no www-site")
    _conditional_print("Unresolvable www-site", "with an unresolvable www-site hostname")
    _conditional_print("Unaccessible www-site", "with an unaccessible www-site")
    _conditional_print("Diverging www-site", "with a www-site different betwwen the Index and makefile")
    _conditional_print("Marked as BROKEN", "with a BROKEN mark (info)")
    _conditional_print("Marked as BROKEN for too long", f"with a BROKEN mark older than {limits['BROKEN since']} days (warning)")
    _conditional_print("Marked as DEPRECATED", "with a DEPRECATED mark (info)")
    _conditional_print("Marked as DEPRECATED for too long", f"with a DEPRECATED mark older than {limits['DEPRECATED since']} days (warning)")
    _conditional_print("Marked as FORBIDDEN", "with a FORBIDDEN mark (info)")
    _conditional_print("Marked as FORBIDDEN for too long", f"with a FORBIDDEN mark older than {limits['FORBIDDEN since']} days (warning)")
    _conditional_print("Marked as IGNORE", "with a IGNORE mark in some cases (info)")
    _conditional_print("Marked as RESTRICTED", "with a RESTRICTED mark (info)")
    _conditional_print("Marked with EXPIRATION_DATE", "with an EXPIRATION_DATE mark (warning)")
    _conditional_print("Unchanged for a long time", f"with a last modification older than {limits['Unchanged since']} days (info)")
    _conditional_print("Both PORTVERSION and DISTVERSION", "with both PORTVERSION and DISTVERSION")
    _conditional_print("Vulnerable port version", "with a vulnerable version (warning)")



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

# Global dictionary of counters:
counters = {
    "FreeBSD ports": 0,
    "Selected ports": 0,
    "Nonexistent Makefile": 0,
    "Nonexistent port-path": 0,
    "Unusual installation-prefix": 0,
    "Too long comments": 0,
    "Uncapitalized comments": 0,
    "Dot-ended comments": 0,
    "Diverging comments": 0,
    "Nonexistent description-file": 0,
    "URL ending description-file": 0,
    "description-file same as comment": 0,
    "Too short description-file": 0,
    "Nonexistent pkg-plist": 0,
    "PLIST_FILES abuse": 0,
    "Diverging maintainers": 0,
    "Unofficial categories": 0,
    "Diverging categories": 0,
    "Empty www-site": 0,
    "Unresolvable www-site": 0,
    "Unaccessible www-site": 0,
    "Diverging www-site": 0,
    "Marked as BROKEN": 0,
    "Marked as BROKEN for too long": 0,
    "Marked as DEPRECATED": 0,
    "Marked as DEPRECATED for too long": 0,
    "Marked as FORBIDDEN": 0,
    "Marked as FORBIDDEN for too long": 0,
    "Marked as IGNORE": 0,
    "Marked as RESTRICTED": 0,
    "Marked with EXPIRATION_DATE": 0,
    "Unchanged for a long time": 0,
    "Both PORTVERSION and DISTVERSION": 0,
    "Vulnerable port version": 0,
    "Skipped vulnerability checks": 0,
}

# Global dictionary of notifications to port maintainers:
notifications = {}

####################################################################################################
def notify_maintainer(maintainer, error, port):
    """ Notify a maintainer about an error related to a port """
    if maintainer in notifications:
        if error in notifications[maintainer]:
            if port not in notifications[maintainer][error]:
                notifications[maintainer][error].append(port)
        else:
            notifications[maintainer][error] = [port]
    else:
        notifications[maintainer] = {error: [port]}



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import re
import socket
import urllib.request

from .library import counters, notify_maintainer

# Headers and timeout delay for HTTP(S) requests:
HTTP_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language": "en;q=1.0, en-US;q=0.8, *;q=0.5",
    "Accept-Encoding": "identity",
    "Connection": "keep-alive",
}
CONNECTION_TIMEOUT = 10 # seconds

####################################################################################################
def _resolve_hostname(hostname):
    """ Resolve the hostname into an IP address or raise a NameError exception """
    try:
        ip_address = socket.gethostbyname(hostname)
    except socket.gaierror as error:
        error_message = re.sub(r".*] ", "", str(error))
        raise NameError(error_message) from error

    return ip_address


####################################################################################################
def _handle_url_errors(port_name, www_site, error, maintainer):
    """ Decides what to do with the multiple possible fetching errors """
    is_unaccessible = False
    reason = "Unaccessible www-site"
    if error.lower().startswith("http error"):
        error_code = re.sub(r"http error ","", error.lower())
        error_message = re.sub(r".*: *", "", error_code)
        error_code = re.sub(r":.*","", error_code)
        if error_code == "404":
            reason = "HTTP Error 404 (Not found) on www-site"
            logging.error("%s '%s' for port %s", reason, www_site, port_name)
            is_unaccessible = True
        elif error_code == "410":
            reason = "HTTP Error 410 (Gone) on www-site"
            logging.error("%s '%s' for port %s", reason, www_site, port_name)
            is_unaccessible = True
        elif error_code == "401":
            reason = "HTTP Error 401 (Unauthorized) on www-site"
            logging.error("%s '%s' for port %s", reason, www_site, port_name)
            is_unaccessible = True
        elif error_code == "403":
            reason = "HTTP Error 403 (Forbidden) on www-site"
            logging.error("%s '%s' for port %s", reason, www_site, port_name)
            is_unaccessible = True
        elif error_code == "451":
            reason = "HTTP Error 451 (Unavailable for legal reasons) on www-site"
            logging.error("%s '%s' for port %s", reason, www_site, port_name)
            is_unaccessible = True
        elif error_code == "301":
            reason = "HTTP Error 301 (Moved permanently) on www-site"
            logging.warning("%s '%s' for port %s", reason, www_site, port_name)
        elif error_code == "308":
            reason = "HTTP Error 308 (Permanent redirect) on www-site"
            logging.warning("%s '%s' for port %s", reason, www_site, port_name)
        else:
            # we don't consider 3xx, 5xx, and remaining 4xx errors to be a reliable sign
            # of a definitive www-site issue
            reason = "HTTP Error " + error_code + " (" + error_message + ") on www-site"
            logging.warning("%s '%s' for port %s", reason, www_site, port_name)
    elif error == "<urlopen error [Errno 8] Name does not resolve>":
        reason = "Unresolvable www-site"
        logging.error("%s '%s' for port %s", reason, www_site, port_name)
        counters["Unresolvable www-site"] += 1
    else:
        logging.debug("%s (%s) '%s' for port %s", reason, error, www_site, port_name)

    if is_unaccessible:
        notify_maintainer(maintainer, reason, port_name)

    return is_unaccessible


####################################################################################################
def check_www_site(ports, check_host, check_url):
    """ Checks the www-site field existence
    Rules at https://docs.freebsd.org/en/books/porters-handbook/makefiles/#makefile-www
    """
    unresolvable_hostnames = []
    url_ok = []
    url_ko = {}
    for name, port in ports.items():
        if port["www-site"] == "":
            logging.error("Empty www-site for port %s", name)
            counters["Empty www-site"] += 1
            notify_maintainer(port["maintainer"], "Empty www-site", name)

        elif check_host:
            hostname = re.sub(r"http[s]*://", "", port["www-site"])
            hostname = re.sub(r"/.*", "", hostname)
            hostname = re.sub(r":[0-9]*", "", hostname)
            resolvable = True
            if hostname in unresolvable_hostnames:
                resolvable = False
            else:
                try:
                    _ = _resolve_hostname(hostname)
                except NameError:
                    resolvable = False
                    unresolvable_hostnames.append(hostname)

            if not resolvable:
                logging.error("Unresolvable www-site '%s' for port %s", hostname, name)
                counters["Unresolvable www-site"] += 1
                notify_maintainer(port["maintainer"], "Unresolvable www-site", name)

            elif port["www-site"] in url_ok:
                pass

            elif port["www-site"] in url_ko:
                if _handle_url_errors(name, port["www-site"], url_ko[port["www-site"]], port["maintainer"]):
                    counters["Unaccessible www-site"] += 1

            elif check_url:
                request = urllib.request.Request(port["www-site"], headers=HTTP_HEADERS)
                try:
                    with urllib.request.urlopen(request, timeout=CONNECTION_TIMEOUT) as http:
                        _ = http.read()
                    url_ok.append(port["www-site"])
                except Exception as error:
                    url_ko[port["www-site"]] = str(error)
                    if _handle_url_errors(name, port["www-site"], str(error), port["maintainer"]):
                        counters["Unaccessible www-site"] += 1

        if "WWW" in port:
            if '$' in port["WWW"]:
                continue # don't try to resolve embedded variables. Ignore check

            if port["www-site"] not in port["WWW"].split():
                logging.error("Diverging www-site between Index and Makefile for port %s", name)
                logging.error("... Index:www-site '%s'", port["www-site"])
                logging.error("... Makefile:WWW   '%s'", port["WWW"])
                counters["Diverging www-site"] += 1
                notify_maintainer(port["maintainer"], "Diverging www-site", name)

    logging.info("Found %d ports with empty www-site", counters["Empty www-site"])
    if check_host:
        logging.info("Found %d ports with unresolvable www-site", counters["Unresolvable www-site"])
        if check_url:
            logging.info("Found %d ports with unaccessible www-site", counters["Unaccessible www-site"])
    logging.info("Found %d ports with diverging www-site", counters["Diverging www-site"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import textwrap

####################################################################################################
def show_categories(ports):
    """ Pretty prints port categories """
    categories = {}
    for _, port in ports.items():
        for category in port["categories"].split():
            if category in categories:
                categories[category] += 1
            else:
                categories[category] = 1

    sorted_categories = dict(sorted(categories.items()))
    all_categories = ""
    for category, count in sorted_categories.items():
        all_categories += f"{category}({count}), "
    all_categories = all_categories[:-2]

    print(f"Showing {len(categories)} categories with ports counts:\n")
    for line in textwrap.wrap(all_categories, width=80, break_on_hyphens=False):
        print(line)



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import os

from .library import counters, notify_maintainer

####################################################################################################
def check_plist(ports, plist_abuse, ports_dir):
    """ Checks the package list existence and compliance with rules
    Rules at https://docs.freebsd.org/en/books/porters-handbook/book/#porting-pkg-plist
    """
    for name, port in ports.items():
        # Use the PORTSDIR we have been told to, rather than the system's one
        port_path = port["port-path"].replace("/usr/ports", ports_dir)

        if os.path.isdir(port_path):
            if not os.path.isfile(port_path + os.sep + "pkg-plist"):
                if not "PLIST_FILES" in port:
                    if not "PLIST" in port and not "PLIST_SUB" in port:
                        logging.debug("Nonexistent pkg-plist/PLIST_FILES/PLIST/PLIST_SUB for port %s", name)
                        counters["Nonexistent pkg-plist"] += 1
                        # Don't notify maintainers because there are too many cases I don't understand!
                else:
                    plist_entries = len(port["PLIST_FILES"].split())
                    if plist_entries >= plist_abuse:
                        logging.warning("PLIST_FILES abuse at %d entries for port %s", plist_entries, name)
                        counters["PLIST_FILES abuse"] += 1
                        notify_maintainer(port["maintainer"], "PLIST_FILES abuse", name)

    logging.info("Found %d ports with nonexistent pkg-plist (use --debug to list them)", counters["Nonexistent pkg-plist"])
    logging.info("Found %d ports with PLIST_FILES abuse", counters["PLIST_FILES abuse"])



#!/usr/bin/env python3
""" portstreelint - FreeBSD ports tree lint
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import os

from .library import counters, notify_maintainer

####################################################################################################
def check_description_file(ports, ports_dir):
    """ Checks the description-file field consistency and existence """
    for name, port in ports.items():
        # Use the PORTSDIR we have been told to, rather than the system's one
        port_path = port["port-path"].replace("/usr/ports", ports_dir)
        description_file = port["description-file"].replace("/usr/ports", ports_dir)

        nonexistent = False
        if not description_file.startswith(port_path):
            if not os.path.isfile(description_file):
                nonexistent = True
        elif not os.path.isdir(port_path):
            pass # already reported
        elif not os.path.isfile(description_file):
            nonexistent = True

        if nonexistent:
            logging.error("Nonexistent description-file '%s' for port %s", description_file, name)
            counters["Nonexistent description-file"] += 1
            notify_maintainer(port["maintainer"], "Nonexistent description-file", name)
        else:
            try:
                with open(description_file, encoding="utf-8", errors="ignore") as file:
                    lines = file.read().splitlines()
            except:
                lines = []

            if lines:
                if lines[-1].strip().startswith("https://") or lines[-1].strip().startswith("http://"):
                    logging.error("URL '%s' ending description-file for port %s", lines[-1].strip(), name)
                    counters["URL ending description-file"] += 1
                    notify_maintainer(port["maintainer"], "URL ending description-file", name)
                    del lines[-1]

                text = " ".join(lines)
                text = text.strip()
                if port["comment"] == text:
                    logging.error("description-file content is identical to comment for port %s", name)
                    counters["description-file same as comment"] += 1
                    notify_maintainer(port["maintainer"], "description-file same as comment", name)
                elif len(text) <= len(port["comment"]):
                    logging.error("description-file content is no longer than comment for port %s", name)
                    counters["Too short description-file"] += 1
                    notify_maintainer(port["maintainer"], "Too short description-file", name)

    logging.info("Found %d ports with nonexistent description-file", counters["Nonexistent description-file"])
    logging.info("Found %d ports with URL ending description-file", counters["URL ending description-file"])
    logging.info("Found %d ports with description-file identical to comment", counters["description-file same as comment"])
    logging.info("Found %d ports with too short description-file", counters["Too short description-file"])
